"""An executor which limits the number of running futures."""
from concurrent.futures import Executor
from threading import Event, Thread, Lock, Semaphore
from collections import namedtuple, deque
import logging

from more_executors._common import _MAX_TIMEOUT
from more_executors.map import _MapFuture

_LOG = logging.getLogger('ThrottleExecutor')

__pdoc__ = {}
__pdoc__['ThrottleExecutor.submit'] = None
__pdoc__['ThrottleExecutor.map'] = None
__pdoc__['ThrottleExecutor.shutdown'] = None


class _ThrottleFuture(_MapFuture):
    def __init__(self, on_cancel):
        self._me_on_cancel = on_cancel
        super(_ThrottleFuture, self).__init__(delegate=None, map_fn=lambda x: x)

    def _me_cancel(self):
        if self._delegate:
            return self._delegate.cancel()
        return self._me_on_cancel(self)


_ThrottleJob = namedtuple('_ThrottleJob', ['future', 'fn', 'args', 'kwargs'])


class ThrottleExecutor(Executor):
    """An `Executor` which delegates to another `Executor` while enforcing
    a limit on the number of futures running concurrently.

    - Callables are submitted to the delegate executor, from a different
      thread than the calling thread.

    - Where `count` is used to initialize this executor, if there
      are already `count` futures submitted to the delegate executor and not
      yet `done()`, additional callables will be queued and only submitted
      to the delegate executor once there are less than `count` futures
      in progress.

    *Since version 1.9.0*
    """
    def __init__(self, delegate, count):
        """Create a new executor.

        - `delegate`: `Executor` instance to which callables will be submitted
        - `count`: maximum number of concurrently running futures
        """
        self._delegate = delegate
        self._to_submit = deque()
        self._lock = Lock()
        self._event = Event()
        self._sem = Semaphore(count)
        self._shutdown = False

        self._thread = Thread(name='ThrottleExecutor', target=self._submit_loop)
        self._thread.daemon = True
        self._thread.start()

    def submit(self, fn, *args, **kwargs):
        out = _ThrottleFuture(on_cancel=self._do_cancel)
        job = _ThrottleJob(out, fn, args, kwargs)
        with self._lock:
            self._to_submit.append(job)
            _LOG.debug("Enqueued: %s", job)
        self._event.set()
        return out

    def shutdown(self, wait=True):
        _LOG.debug("Shutting down")
        self._shutdown = True
        self._delegate.shutdown(wait)
        self._event.set()
        if wait:
            self._thread.join(_MAX_TIMEOUT)

    def _submit_loop(self):
        while not self._shutdown:
            self._event.clear()

            to_submit = []
            with self._lock:
                while self._to_submit:
                    if not self._sem.acquire(False):
                        _LOG.debug("Throttled")
                        break
                    job = self._to_submit.popleft()
                    _LOG.debug("Will submit: %s", job)
                    to_submit.append(job)

                _LOG.debug("Submitting %s, throttling %s", len(to_submit), len(self._to_submit))

            for job in to_submit:
                self._do_submit(job)

            self._event.wait()

    def _do_submit(self, job):
        delegate_future = self._delegate.submit(job.fn, *job.args, **job.kwargs)
        _LOG.debug("Submitted %s yielding %s", job, delegate_future)

        delegate_future.add_done_callback(self._delegate_future_done)
        job.future._set_delegate(delegate_future)

    def _do_cancel(self, future):
        with self._lock:
            for job in self._to_submit:
                if job.future is future:
                    self._to_submit.remove(job)
                    _LOG.debug("Cancelled %s", job)
                    return True
        _LOG.debug("Could not find for cancel: %s", future)
        return False

    def _delegate_future_done(self, future):
        # Whenever an inner future completes, one more execution slot becomes
        # available, and the thread should wake up in case there's something to
        # be submitted
        _LOG.debug("Delegate future done: %s", future)
        self._sem.release()
        self._event.set()



"""Cancel futures when executor is shut down."""
from concurrent.futures import Executor
from threading import RLock
import logging

__pdoc__ = {}
__pdoc__['CancelOnShutdownExecutor.submit'] = None
__pdoc__['CancelOnShutdownExecutor.map'] = None

_LOG = logging.getLogger('CancelOnShutdownExecutor')


class CancelOnShutdownExecutor(Executor):
    """An `Executor` which delegates to another `Executor` and cancels all
    futures when the executor is shut down.

    This class is useful in conjunction with executors having custom cancel
    behavior, such as `more_executors.poll.PollExecutor`.
    """

    def __init__(self, delegate):
        """Create a new executor.

        - `delegate`: `Executor` instance to which callables will be submitted
        """
        self._delegate = delegate
        self._futures = set()
        self._lock = RLock()
        self._shutdown = False

    def shutdown(self, wait=True):
        """Shut down the executor.

        All futures created by this executor which have not yet been completed
        will have `cancel` invoked.  Note that there is no guarantee that the
        cancel will succeed.
        """
        with self._lock:
            if self._shutdown:
                return
            self._shutdown = True

            for f in self._futures.copy():
                cancel = f.cancel()
                _LOG.debug("Cancel %s: %s", f, cancel)

            self._delegate.shutdown(wait)

    def submit(self, fn, *args, **kwargs):
        with self._lock:
            if self._shutdown:
                raise RuntimeError('Cannot submit after shutdown')
            future = self._delegate.submit(fn, *args, **kwargs)
            self._futures.add(future)
            future.add_done_callback(self._futures.discard)
        return future



"""Create futures which must complete within a timeout or be cancelled."""
from concurrent.futures import Executor
from threading import Event, Thread, Lock
from collections import namedtuple
import logging

from monotonic import monotonic

from more_executors.map import _MapFuture
from more_executors._common import _MAX_TIMEOUT

__pdoc__ = {}
__pdoc__['TimeoutExecutor.map'] = None
__pdoc__['TimeoutExecutor.shutdown'] = None
__pdoc__['TimeoutExecutor.submit'] = None

_LOG = logging.getLogger('TimeoutExecutor')

_Job = namedtuple('_Job', ['future', 'delegate_future', 'deadline'])


class _TimeoutFuture(_MapFuture):
    def __init__(self, delegate):
        super(_TimeoutFuture, self).__init__(delegate, lambda x: x)


class TimeoutExecutor(Executor):
    """An `Executor` which delegates to another `Executor` while applying
    a timeout to each returned future.

    For any futures returned by this executor, if the future hasn't
    completed approximately within `timeout` seconds of its creation,
    an attempt will be made to cancel the future.

    Note that only a single attempt is made to cancel any future, and there
    is no guarantee that this will succeed.

    *Since version 1.7.0*
    """
    def __init__(self, delegate, timeout):
        """Create a new executor.

        - `delegate`: the delegate executor to which callables are submitted.
        - `timeout`: timeout (in seconds) after which `future.cancel()` will be
                     invoked on any generated future which has not completed.
        """
        self._delegate = delegate
        self._timeout = timeout
        self._shutdown = False
        self._jobs = []
        self._jobs_lock = Lock()
        self._jobs_write = Event()
        self._job_thread = Thread(
            name='TimeoutExecutor', target=self._job_loop)
        self._job_thread.daemon = True
        self._job_thread.start()

    def submit(self, fn, *args, **kwargs):
        delegate_future = self._delegate.submit(fn, *args, **kwargs)
        future = _TimeoutFuture(delegate_future)
        future.add_done_callback(self._on_future_done)
        job = _Job(future, delegate_future, monotonic() + self._timeout)
        with self._jobs_lock:
            self._jobs.append(job)
        self._jobs_write.set()
        return future

    def shutdown(self, wait=True):
        _LOG.debug("shutdown")
        self._shutdown = True
        self._jobs_write.set()
        self._delegate.shutdown(wait)
        if wait:
            self._job_thread.join(_MAX_TIMEOUT)

    def _partition_jobs(self):
        pending = []
        overdue = []
        now = monotonic()
        for job in self._jobs:
            if job.future.done():
                _LOG.debug("Discarding job for completed future: %s", job)
            elif job.deadline < now:
                overdue.append(job)
            else:
                pending.append(job)
        return (pending, overdue)

    def _on_future_done(self, future):
        _LOG.debug("Waking thread for %s", future)
        self._jobs_write.set()

    def _do_cancel(self, job):
        _LOG.debug("Attempting cancel: %s", job)
        cancel_result = job.future.cancel()
        _LOG.debug("Cancel of %s resulted in %s", job, cancel_result)

    def _job_loop(self):
        while not self._shutdown:
            _LOG.debug("job loop")
            self._jobs_write.clear()

            with self._jobs_lock:
                (pending, overdue) = self._partition_jobs()
                self._jobs = pending

            _LOG.debug("jobs: %s overdue, %s pending", len(overdue), len(pending))

            for job in overdue:
                self._do_cancel(job)

            wait_time = None
            if pending:
                earliest = min([job.deadline for job in pending])
                wait_time = max(earliest - monotonic(), 0)

            _LOG.debug("Wait until %s", wait_time)
            self._jobs_write.wait(wait_time)
